use crate::ast::named::Named;
use crate::ast::*;
use crate::name_analysis::fqn_context::FqnContext;
use crate::name_analysis::symbol::{
    ClassSymbol, FunctionSymbol, ModuleSymbol, SourceDefinition, Symbol, VariableSymbol,
};
use crate::name_analysis::symbol_table::{SymbolInsertError, SymbolTable};
use crate::name_analysis::DiagnosticsContainer;
use codespan_reporting::diagnostic::{Diagnostic, Label};
use std::range::Range;

fn handle_insert_error(
    err: SymbolInsertError,
    error_symbol_name: &str,
    error_file_id: usize,
    error_range: Range<usize>,
    symbol_types: &str,
    diagnostics: &mut DiagnosticsContainer,
) {
    match err {
        SymbolInsertError::SymbolAlreadyDefined(s) => {
            let already_defined_definition = s.definition();

            diagnostics.add(
                Diagnostic::error()
                    .with_message(format!(
                        "{} symbol '{}' already defined in the current scope.",
                        symbol_types, error_symbol_name,
                    ))
                    .with_label(
                        Label::primary(error_file_id, error_range)
                            .with_message("Symbol duplicated here."),
                    )
                    .with_label(
                        Label::secondary(
                            already_defined_definition.file_id(),
                            already_defined_definition.range(),
                        )
                        .with_message("Symbol defined here."),
                    ),
            );
        }
    }
}

fn gather_fully_qualified_name(
    fully_qualified_name: &mut FullyQualifiedName,
    symbol_table: &mut SymbolTable,
) {
    fully_qualified_name.set_scope_id(symbol_table.current_scope_id());
}

fn gather_type_use(
    type_use: &mut TypeUse,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    match type_use {
        TypeUse::Void => {}
        TypeUse::InterfaceOrClass(interface_or_class_type_use) => {
            gather_interface_or_class_type_use(
                interface_or_class_type_use,
                symbol_table,
                fqn_context,
                diagnostics,
            );
        }
        TypeUse::Tuple(tuple_type_use) => {
            gather_tuple_type_use(tuple_type_use, symbol_table, fqn_context, diagnostics);
        }
        TypeUse::Function(function_type_use) => {
            gather_function_type_use(function_type_use, symbol_table, fqn_context, diagnostics);
        }
    }
}

fn gather_interface_or_class_type_use(
    interface_or_class_type_use: &mut InterfaceOrClassTypeUse,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_fully_qualified_name(&mut interface_or_class_type_use.fqn, symbol_table);
    gather_generic_arguments(
        &mut interface_or_class_type_use.generics,
        symbol_table,
        fqn_context,
        diagnostics,
    );
}

fn gather_tuple_type_use(
    tuple_type_use: &mut TupleTypeUse,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for generic_argument in &mut tuple_type_use.arguments.0 {
        gather_type_use(generic_argument, symbol_table, fqn_context, diagnostics);
    }
}

fn gather_function_type_use(
    function_type_use: &mut FunctionTypeUse,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    todo!()
}

fn gather_generic_arguments(
    generic_arguments: &mut GenericArguments,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for argument in &mut generic_arguments.0 {
        gather_type_use(argument, symbol_table, fqn_context, diagnostics);
    }
}

pub(super) fn gather_module_level_declaration(
    declaration: &mut ModuleLevelDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use ModuleLevelDeclaration::*;
    match declaration {
        Module(module_declaration) => {
            gather_module_declaration(module_declaration, symbol_table, fqn_context, diagnostics);
        }
        Class(class_declaration) => {
            gather_class_declaration(class_declaration, symbol_table, fqn_context, diagnostics)
        }
        Function(function_definition) => {
            gather_function_definition(function_definition, symbol_table, fqn_context, diagnostics)
        }
        PlatformFunction(platform_function_definition) => {
            gather_platform_function_definition(
                platform_function_definition,
                symbol_table,
                fqn_context,
                diagnostics,
            );
        }
        _ => todo!(),
    }
}

fn gather_module_declaration(
    declaration: &mut ModuleDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    // 1. Add mod identifier symbol
    // 2. Update fqn context
    // 3. Push scope
    // 4. Process declarations
    // 5. Pop scope

    let module_name = declaration.identifier.name();

    let insert_result = symbol_table.insert(
        &module_name,
        Symbol::Module(ModuleSymbol::new(
            &fqn_context.resolve(&module_name),
            &module_name,
            declaration.is_public,
            &declaration.identifier,
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &module_name,
            declaration.identifier.file_id,
            declaration.identifier.range,
            "module/type",
            diagnostics,
        )
    }

    fqn_context.push(module_name.to_string());

    symbol_table.push_scope(&format!("Module '{}' Scope", module_name));
    for inner_declaration in &mut declaration.declarations {
        gather_module_level_declaration(inner_declaration, symbol_table, fqn_context, diagnostics);
    }
    symbol_table.pop_scope()
}

fn gather_class_declaration(
    class_declaration: &mut ClassDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let declared_name = class_declaration.identifier.name();
    let resolved_name = fqn_context.resolve(&declared_name);

    let insert_result = symbol_table.insert(
        &declared_name,
        Symbol::Class(ClassSymbol::new(
            &resolved_name,
            &declared_name,
            class_declaration.is_public,
            &class_declaration.identifier,
        )),
    );
    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &declared_name,
            class_declaration.identifier.file_id,
            class_declaration.identifier.range,
            "interface/class",
            diagnostics,
        );
    }

    // todo: scopes, generics, etc.
}

fn gather_function_definition(
    function: &mut FunctionDefinition,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let declared_name = function.identifier.name();
    let resolved_name = fqn_context.resolve(&declared_name);

    let insert_result = symbol_table.insert(
        &declared_name,
        Symbol::Function(FunctionSymbol::new(
            &resolved_name,
            &declared_name,
            function.is_public,
            false,
            &function.identifier,
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &declared_name,
            function.identifier.file_id,
            function.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    function
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    symbol_table.push_scope(&format!("FunctionScope({})", resolved_name));
    gather_parameters(
        &mut function.parameters,
        symbol_table,
        fqn_context,
        diagnostics,
    );
    gather_function_body(&mut function.body, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_platform_function_definition(
    platform_function_declaration: &mut PlatformFunctionDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let declared_name = platform_function_declaration.identifier.name();
    let fully_qualified_name = fqn_context.resolve(&declared_name);

    let insert_result = symbol_table.insert(
        &declared_name,
        Symbol::Function(FunctionSymbol::new(
            &fully_qualified_name,
            &declared_name,
            platform_function_declaration.is_public,
            true,
            &platform_function_declaration.identifier,
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &declared_name,
            platform_function_declaration.identifier.file_id,
            platform_function_declaration.identifier.range,
            "(Platform-) Function",
            diagnostics,
        );
    }

    let declared_name_as_string = platform_function_declaration.identifier.name().to_string();

    platform_function_declaration
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    symbol_table.push_scope(&format!("FunctionScope({})", declared_name_as_string));
    gather_parameters(
        &mut platform_function_declaration.parameters,
        symbol_table,
        fqn_context,
        diagnostics,
    );
    symbol_table.pop_scope();
}

fn gather_parameters(
    parameters: &mut Parameters,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for parameter in &mut parameters.0 {
        gather_parameter(parameter, symbol_table, fqn_context, diagnostics);
    }
}

fn gather_parameter(
    parameter: &mut Parameter,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let parameter_name = parameter.identifier.name();

    let insert_result = symbol_table.insert(
        &parameter_name,
        Symbol::Variable(VariableSymbol::new(
            &parameter_name,
            false,
            SourceDefinition::from_identifier(&parameter.identifier),
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &parameter_name,
            parameter.identifier.file_id,
            parameter.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    parameter
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    gather_type_use(
        &mut parameter.type_use,
        symbol_table,
        fqn_context,
        diagnostics,
    )
}

fn gather_function_body(
    function_body: &mut FunctionBody,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::FunctionBody::*;
    match function_body {
        Block(block) => gather_block_statement_inner(block, symbol_table, fqn_context, diagnostics),
        _ => todo!(),
    }
}

fn gather_block_statement(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    symbol_table.push_scope("BlockStatementScope");
    gather_block_statement_inner(block, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_block_statement_inner(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for statement in &mut block.statements {
        gather_statement(statement, symbol_table, fqn_context, diagnostics);
    }
    if let Some(expression) = &mut block.expression {
        gather_expression(expression, symbol_table, diagnostics);
    }
}

fn gather_statement(
    statement: &mut Statement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Statement::*;
    match statement {
        BlockStatement(block) => {
            gather_block_statement(block, symbol_table, fqn_context, diagnostics)
        }
        VariableDeclarationStatement(variable_declaration) => {
            gather_variable_declaration(variable_declaration, symbol_table, diagnostics)
        }
        AssignStatement(assign_statement) => {
            gather_assign_statement(assign_statement, symbol_table, fqn_context, diagnostics)
        }
        CallStatement(call_statement) => {
            gather_call_statement(call_statement, symbol_table, diagnostics)
        }
        _ => todo!(),
    }
}

fn gather_variable_declaration(
    variable_declaration: &mut VariableDeclarationStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    let variable_name = variable_declaration.identifier.name();

    let insert_result = symbol_table.insert(
        &variable_name,
        Symbol::Variable(VariableSymbol::new(
            &variable_name,
            variable_declaration.is_mutable,
            SourceDefinition::from_identifier(&variable_declaration.identifier),
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &variable_name,
            variable_declaration.identifier.file_id,
            variable_declaration.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    variable_declaration
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    if let Some(initializer) = &mut variable_declaration.initializer {
        gather_expression(initializer, symbol_table, diagnostics);
    }
}

fn gather_assign_statement(
    assign_statement: &mut AssignStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut assign_statement.lhs, symbol_table, diagnostics);
    gather_expression(&mut assign_statement.rhs, symbol_table, diagnostics);
}

fn gather_call_statement(
    call_statement: &mut CallStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut call_statement.0, symbol_table, diagnostics);
}

fn gather_expression(
    expression: &mut Expression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Expression::*;
    match expression {
        Ternary(ternary_expression) => {
            gather_ternary_expression(ternary_expression, symbol_table, diagnostics);
        }
        Binary(binary_expression) => {
            gather_binary_expression(binary_expression, symbol_table, diagnostics);
        }
        UnaryPrefix(prefix_expression) => {
            gather_prefix_expression(prefix_expression, symbol_table, diagnostics);
        }
        UnarySuffix(suffix_expression) => {
            gather_suffix_expression(suffix_expression, symbol_table, diagnostics);
        }
        Call(call_expression) => {
            gather_call_expression(call_expression, symbol_table, diagnostics);
        }
        ObjectAccess(object_access) => {
            gather_object_access(object_access, symbol_table, diagnostics);
        }
        Literal(literal) => {
            gather_literal(literal, symbol_table, diagnostics);
        }
        FullyQualifiedName(fully_qualified_name) => {
            gather_fully_qualified_name(fully_qualified_name, symbol_table);
        }
        Closure(closure) => {
            gather_closure(closure, symbol_table, diagnostics);
        }
    }
}

fn gather_ternary_expression(
    ternary_expression: &mut TernaryExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut ternary_expression.condition, symbol_table, diagnostics);
    gather_expression(
        &mut ternary_expression.true_expression,
        symbol_table,
        diagnostics,
    );
    gather_expression(
        &mut ternary_expression.false_expression,
        symbol_table,
        diagnostics,
    );
}

fn gather_binary_expression(
    binary_expression: &mut BinaryExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut binary_expression.left, symbol_table, diagnostics);
    gather_expression(&mut binary_expression.right, symbol_table, diagnostics);
}

fn gather_prefix_expression(
    prefix_expression: &mut PrefixExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut prefix_expression.expression, symbol_table, diagnostics);
}

fn gather_suffix_expression(
    suffix_expression: &mut SuffixExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut suffix_expression.expression, symbol_table, diagnostics);
}

fn gather_call_expression(
    call_expression: &mut CallExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut call_expression.callee, symbol_table, diagnostics);
    if let Some(turbo_fish) = &mut call_expression.turbo_fish {
        gather_turbo_fish(turbo_fish, symbol_table, diagnostics);
    }
    for call_argument in &mut call_expression.arguments.0 {
        gather_expression(&mut call_argument.0, symbol_table, diagnostics);
    }
}

fn gather_turbo_fish(
    turbo_fish: &mut TurboFish,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    todo!()
}

fn gather_closure(
    closure: &mut Closure,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    todo!()
}

fn gather_object_access(
    object_access: &mut ObjectAccess,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut object_access.receiver, symbol_table, diagnostics);
    for object_navigation in &mut object_access.navigations.0 {
        match object_navigation {
            ObjectNavigation::Index(index_expression) => {
                gather_expression(index_expression, symbol_table, diagnostics);
            }
            ObjectNavigation::Identifier(identifier) => {
                identifier.set_scope_id(symbol_table.current_scope_id());
            }
        }
    }
}

fn gather_literal(
    literal: &mut Literal,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    match literal {
        Literal::DString(d_string) => todo!(),
        Literal::BacktickString(backtick_string) => todo!(),
        _ => {}
    }
}
